==> octosql-0.6.2.tar.gz: octosql-0.6.2/execution/nodes/distinct.go <==
package nodes

import (
	"fmt"

	"github.com/google/btree"

	. "github.com/cube2222/octosql/execution"
	"github.com/cube2222/octosql/octosql"
)

type Distinct struct {
	source Node
}

func NewDistinct(source Node) *Distinct {
	return &Distinct{
		source: source,
	}
}

type distinctItem struct {
	Values []octosql.Value
	Count  int
}

func (item *distinctItem) Less(than btree.Item) bool {
	thanTyped, ok := than.(*distinctItem)
	if !ok {
		panic(fmt.Sprintf("invalid distinct key comparison: %T", than))
	}

	for i := 0; i < len(item.Values); i++ {
		if comp := item.Values[i].Compare(thanTyped.Values[i]); comp != 0 {
			return comp == -1
		}
	}

	return false
}

func (o *Distinct) Run(execCtx ExecutionContext, produce ProduceFn, metaSend MetaSendFn) error {
	recordCounts := btree.New(BTreeDefaultDegree)
	o.source.Run(
		execCtx,
		func(ctx ProduceContext, record Record) error {
			item := recordCounts.Get(&distinctItem{Values: record.Values})
			var itemTyped *distinctItem
			if item == nil {
				itemTyped = &distinctItem{
					Values: record.Values,
					Count:  0,
				}
			} else {
				var ok bool
				itemTyped, ok = item.(*distinctItem)
				if !ok {
					panic(fmt.Sprintf("invalid order by item: %v", item))
				}
			}
			if !record.Retraction {
				itemTyped.Count++
			} else {
				itemTyped.Count--
			}
			if itemTyped.Count > 0 {
				// New record.
				if !record.Retraction && itemTyped.Count == 1 {
					if err := produce(ctx, record); err != nil {
						return fmt.Errorf("couldn't produce new record: %w", err)
					}
				}
				recordCounts.ReplaceOrInsert(itemTyped)
			} else {
				if err := produce(ctx, record); err != nil {
					return fmt.Errorf("couldn't retract record record: %w", err)
				}
				recordCounts.Delete(itemTyped)
			}
			return nil
		},
		func(ctx ProduceContext, msg MetadataMessage) error {
			return nil
		},
	)

	return nil
}

==> octosql-0.7.2.tar.gz: octosql-0.7.2/execution/nodes/distinct.go <==
package nodes

import (
	"fmt"

	"github.com/tidwall/btree"

	. "github.com/cube2222/octosql/execution"
	"github.com/cube2222/octosql/octosql"
)

type Distinct struct {
	source Node
}

func NewDistinct(source Node) *Distinct {
	return &Distinct{
		source: source,
	}
}

type distinctItem struct {
	Values []octosql.Value
	Count  int
}

func (o *Distinct) Run(execCtx ExecutionContext, produce ProduceFn, metaSend MetaSendFn) error {
	recordCounts := btree.NewGenericOptions(func(item, than *distinctItem) bool {
		for i := 0; i < len(item.Values); i++ {
			if comp := item.Values[i].Compare(than.Values[i]); comp != 0 {
				return comp == -1
			}
		}

		return false
	}, btree.Options{
		NoLocks: true,
	})
	o.source.Run(
		execCtx,
		func(ctx ProduceContext, record Record) error {
			item, ok := recordCounts.Get(&distinctItem{Values: record.Values})
			if !ok {
				item = &distinctItem{
					Values: record.Values,
					Count:  0,
				}
			}
			if !record.Retraction {
				item.Count++
			} else {
				item.Count--
			}
			if item.Count > 0 {
				// New record.
				if !record.Retraction && item.Count == 1 {
					if err := produce(ctx, record); err != nil {
						return fmt.Errorf("couldn't produce new record: %w", err)
					}
					recordCounts.Set(item)
				}
			} else {
				if err := produce(ctx, record); err != nil {
					return fmt.Errorf("couldn't retract record record: %w", err)
				}
				recordCounts.Delete(item)
			}
			return nil
		},
		func(ctx ProduceContext, msg MetadataMessage) error {
			return nil
		},
	)

	return nil
}
